package token

import "bytes"

// Tokenize turns the file contents into a list of tokens.
func Tokenize(buffer []byte) List {
	var (
		i      Position
		tokens = make(List, 0, 8+len(buffer)/2)
	)

	for i < Position(len(buffer)) {
		switch buffer[i] {
		case ' ', '\t':
		case ',':
			tokens = append(tokens, Token{Kind: Separator, Position: i, Length: 1})
		case '(':
			tokens = append(tokens, Token{Kind: GroupStart, Position: i, Length: 1})
		case ')':
			tokens = append(tokens, Token{Kind: GroupEnd, Position: i, Length: 1})
		case '{':
			tokens = append(tokens, Token{Kind: BlockStart, Position: i, Length: 1})
		case '}':
			tokens = append(tokens, Token{Kind: BlockEnd, Position: i, Length: 1})
		case '[':
			tokens = append(tokens, Token{Kind: ArrayStart, Position: i, Length: 1})
		case ']':
			tokens = append(tokens, Token{Kind: ArrayEnd, Position: i, Length: 1})
		case '\n':
			tokens = append(tokens, Token{Kind: NewLine, Position: i, Length: 1})
		case '/':
			if i+1 >= Position(len(buffer)) || buffer[i+1] != '/' {
				position := i
				i++

				for i < Position(len(buffer)) && isOperator(buffer[i]) {
					i++
				}

				tokens = append(tokens, Token{Kind: Operators[string(buffer[position:i])], Position: position, Length: Length(i - position)})
			} else {
				position := i

				for i < Position(len(buffer)) && buffer[i] != '\n' {
					i++
				}

				tokens = append(tokens, Token{Kind: Comment, Position: position, Length: Length(i - position)})
			}

			continue

		case '"', '\'':
			limiter := buffer[i]
			start := i
			end := Position(len(buffer))
			i++

			for i < Position(len(buffer)) {
				if buffer[i] == limiter {
					end = i + 1
					i++
					break
				}

				i++
			}

			kind := String

			if limiter == '\'' {
				kind = Rune
			}

			tokens = append(tokens, Token{Kind: kind, Position: start, Length: Length(end - start)})
			continue

		default:
			if isIdentifierStart(buffer[i]) {
				position := i
				i++

				for i < Position(len(buffer)) && isIdentifier(buffer[i]) {
					i++
				}

				identifier := buffer[position:i]
				kind := Identifier

				switch identifier[0] {
				case 'i':
					switch {
					case bytes.Equal(identifier, []byte("if")):
						kind = If
					case bytes.Equal(identifier, []byte("import")):
						kind = Import
					}
				case 'l':
					switch {
					case bytes.Equal(identifier, []byte("loop")):
						kind = Loop
					}
				case 'r':
					switch {
					case bytes.Equal(identifier, []byte("return")):
						kind = Return
					}
				}

				tokens = append(tokens, Token{Kind: kind, Position: position, Length: Length(len(identifier))})
				continue
			}

			if isNumber(buffer[i]) {
				position := i
				i++

				for i < Position(len(buffer)) && isNumber(buffer[i]) {
					i++
				}

				tokens = append(tokens, Token{Kind: Number, Position: position, Length: Length(i - position)})
				continue
			}

			if isOperator(buffer[i]) {
				position := i
				i++

				for i < Position(len(buffer)) && isOperator(buffer[i]) {
					i++
				}

				tokens = append(tokens, Token{Kind: Operators[string(buffer[position:i])], Position: position, Length: Length(i - position)})
				continue
			}

			tokens = append(tokens, Token{Kind: Invalid, Position: i, Length: 1})
		}

		i++
	}

	tokens = append(tokens, Token{Kind: EOF, Position: i, Length: 0})
	return tokens
}

func isIdentifier(c byte) bool {
	return isLetter(c) || isNumber(c) || c == '_'
}

func isIdentifierStart(c byte) bool {
	return isLetter(c) || c == '_'
}

func isLetter(c byte) bool {
	return (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z')
}

func isNumber(c byte) bool {
	return (c >= '0' && c <= '9')
}

func isOperator(c byte) bool {
	switch c {
	case '=', ':', '.', '+', '-', '*', '/', '<', '>', '&', '|', '^', '%', '!':
		return true
	default:
		return false
	}
}
